
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# envs.py :: # contains custom made simple gym environments for testing
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
import gym
import gym.spaces
import numpy as np
import matplotlib.pyplot as plt 
from matplotlib.patches import Circle
#import mpl_toolkits.mplot3d.art3d as art3d

class SimpleEnv(gym.Env):
    def __init__(self, 
            discrete_action=True, 
            min_dis_from_goal=2.0, 
            xrad = 15, 
            yrad = 10, 
            alim = 1, 
            min_dis_from_term = 2.0, 
            max_episode_steps = 400,
            seed=None
            ) -> None:
        # alim is usually 1 <--- not a tuple
        super().__init__()
        self.rng = np.random.default_rng(seed)
        self.min_dis_from_goal = min_dis_from_goal
        self.xl, self.xh = -xrad, xrad
        self.yl, self.yh = -xrad, xrad
        self._max_episode_steps=max_episode_steps
        self.min_dis_from_term=min_dis_from_term
        self.discrete_action = discrete_action
        self.space_dtype = (np.int32 if discrete_action else np.float32 )
        self.X = np.array(  [   0, 0   ]   )
        self.base_noise = np.array([ xrad*0.05 , yrad*0.05 ])
        self.observation_space = gym.spaces.Box(
                            #   agent-x,   agent-y     goal-x,      goal-y,                                         
            low = np.array([    self.xl,   self.yl,    self.xl,     self.yl     ]),
            high= np.array([    self.xh,   self.yh,    self.xh,     self.yh     ]),
            dtype=self.space_dtype)
        #dlim = (alim)/(2**0.5)
        if discrete_action:
            self.action_space = gym.spaces.Discrete(8) 
            self.action_mapper = {
                0   :   np.array([0, alim], self.space_dtype), # North

                1   :   np.array([0, -alim], self.space_dtype), # South

                2   :   np.array([alim, 0], self.space_dtype), # East

                3   :   np.array([-alim, 0], self.space_dtype), # West


                4   :   np.array([alim, alim], self.space_dtype), # North E
                5   :   np.array([alim, -alim], self.space_dtype), # South E
                6   :   np.array([-alim, alim], self.space_dtype), # North W
                7   :   np.array([-alim, -alim], self.space_dtype), # South W
            }
            self.act = self.act_discrete
        else:
            self.action_space = gym.spaces.Box(
                    low = np.array([-alim, -alim]),
                    high= np.array([alim, alim]),
                    dtype=self.space_dtype)
            self.act = self.act_continuous
        self.S = None
        self.Si = self.observation_space.sample() 
        self.Si[2:4] = (-self.Si[0:2]) + self.state_noise()
        self.frozen=False
        self.clip=True

    def _distance_from_goal(self):
        return (np.sum((self.S[0:2]-self.S[2:4])**2))**0.5

    def _distance_from_terminator(self):
        return (np.sum((self.S[0:2]-self.X[0:2])**2))**0.5

    def state_noise(self):
        return  (self.base_noise + ( self.rng.random((2,))-0.5 ) ) 
    def reset(self):
        if not self.frozen:
            self.Si = self.observation_space.sample() 
            self.Si[2:4] = (-self.Si[0:2]) + self.state_noise()
        self.S = np.copy(self.Si) 
        self.T=0
        self.dis = self._distance_from_goal()
        self.tis = self._distance_from_terminator()
        self.hist=[self.S.flatten()]
        return self.S.flatten()

    def act(self, action):
        print(f'[WARNING]: Not implemented! act({action = })')
        return None

    def step(self, action):
        self.S[0:2] -= self.act(action)
        self.hist.append(self.S.flatten())
        dis = self._distance_from_goal()
        tis = self._distance_from_terminator()

        reward = (self.dis-dis)
        self.dis=dis
        self.tis=tis
        self.T+=1
        
        if self.observation_space.contains(self.S):
            if self.dis<self.min_dis_from_goal :
                done=True
                reward+=10
            elif tis<self.min_dis_from_term:
                done=True
                reward-=10
            else:
                done = (self.T>=self._max_episode_steps)
        else:
            done=True
            reward-=10
        return self.S.flatten(), reward, done, {'dis': self.dis, 'tis': self.tis}
        
    def render(self, caption=""):
        fig = plt.figure(figsize=(6,6))
        ax = fig.add_subplot()
        ax.set_xlim((self.xl, self.xh))
        ax.set_ylim((self.yl, self.yh))
        if caption:
            fig.suptitle(caption)

        if len(self.hist)>1:
            ax.scatter([self.hist[0][0]], [self.hist[0][1]], marker='s', color='tab:blue')
            #ax.scatter([self.hist[0][2]], [self.hist[0][3]], marker='s', color='tab:red')
            for i in range(1, len(self.hist)):
                ax.annotate("", 
                           xy=( self.hist[i][0], self.hist[i][1] ),
                           xytext=(self.hist[i-1][0], self.hist[i-1][1]), 
                  arrowprops=dict(arrowstyle="->", linewidth=0.8, color='tab:pink'))

        
        ax.scatter([self.S[0]], [self.S[1]], marker='o', color='tab:blue')     # player position
        ax.scatter([self.S[2]], [self.S[3]], marker='s', color='tab:green')      # goal position
        ax.add_patch(Circle(   (self.S[2], self.S[3])   , self.min_dis_from_goal, color='tab:green', linewidth=0.8, fill=False))
        ax.scatter([self.X[0]], [self.X[1]], marker='x', color='tab:red')      # t position
        ax.add_patch(Circle(   (self.X[0], self.X[1])   , self.min_dis_from_term, color='tab:red', linewidth=0.8, fill=False))
        plt.show()

    def act_continuous(self, action):
        action = action.astype(np.float32)
        if self.clip:
            action = np.clip(action, self.action_space.low, self.action_space.high)
        return action

    def act_discrete(self, action):
        if self.clip:
            action = int(abs(action)) % self.action_space.n
        return self.action_mapper[action]



#-----------------------------------------------------------------------------------------------------
# Foot-Note:
""" NOTE:
    * Author:           Nelson.S
"""
#-----------------------------------------------------------------------------------------------------
